package de.bioinf.appl.distg;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JTabbedPane;

import de.bioinf.appl.Appl;
import de.bioinf.appl.ApplParam;
import de.bioinf.ui.Borders;
import de.bioinf.ui.ColorMap;
import de.bioinf.ui.FileSelectorDialog;
import de.bioinf.ui.GridBagPanel;
import de.bioinf.ui.HelpDialog;
import de.bioinf.ui.InputComponent;
import de.bioinf.ui.InputDialog;
import de.bioinf.ui.IntegerTextField;
import de.bioinf.ui.MenuBar;
import de.bioinf.ui.Window;
import de.bioinf.ui.graph.CoordPanel;
import de.bioinf.ui.graph.GraphPanel;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Executor;
import de.bioinf.utils.IntValueRange;
import de.bioinf.utils.Logger;
import de.bioinf.utils.Print;
import de.bioinf.utils.Source;
import de.bioinf.utils.Utils;
import de.bioinf.utils.Www;

/**
 * This is an Application to display the Distribution Graph of data
 * 
 * @author Jan
 */
public class DistGAppl extends Appl {
	private DataView g_data = null;
	private Menu g_menu = null;
	private ParamsDialog g_paramsDialog = null;
	private FileDialog g_fileDialog = null;
	private ExportDialog g_exportDialog = null;
	private HelpDialog g_helpDialog = null;

	private ColorMap colormap = null;
	private ArrayList<DistGValues> valuelist = new ArrayList<DistGValues>();
	private DistGParams params = new DistGParams();

	// **
	// Executors
	// **

	class LoadFileExecutor extends Executor {
		Source source = null;
		// The range of lengths
		IntValueRange range = null;

		public LoadFileExecutor(Source source, IntValueRange range) {
			super();
			this.source = source;
			this.range = range;
		}

		@Override
		public void execute() throws BioinfException {
			BioinfException.fire(!DistGValues.isLoaded(source, valuelist), "Cannot load a file twice!");

			try {
				DistGValues values = DistGValues.load(source, params);
				valuelist.add(values);
				colormap.put(source.getLocation(), null);
				g_menu.addHistogram(values);

				// params.range.add(values.counts.length);
				// g_paramsDialog.init(params);

				if (valuelist.size() == 1) {
					g_data.init();
				} else {
					g_data.add(valuelist.size() - 1);
				}
			} catch (Exception e) {
				throw BioinfException.convert(e);
			}
		}
	}

	/**
	 * Save data to a local file.
	 */
	class SaveDataExecutor extends Executor {
		String filename = null;
		boolean all = true;

		public SaveDataExecutor(String filename, boolean all) {
			super();
			this.filename = filename;
			this.all = all;
		}

		@Override
		public void execute() throws BioinfException {
			g_data.saveValues(filename, all);
		}
	}

	// **
	// Dialogs
	// **

	class Menu extends MenuBar implements ActionListener {
		MenuItem m_loadFile = null;
		MenuItem m_saveAll = null;
		MenuItem m_saveGraph = null;
		MenuItem m_params = null;
		MenuItem m_colorize = null;
		MenuItem m_decolorize = null;
		MenuItem m_clear = null;
		MenuItem m_print = null;
		MenuItem m_export = null;
		MenuItem m_exit = null;
		MenuItem m_helpUsage = null;

		class MenuItem extends JMenuItem {
			public MenuItem(String title) {
				super(title);
				addActionListener(Menu.this);
			}
		}

		class HistItem extends JMenuItem implements ActionListener {
			DistGValues values = null;

			public HistItem(DistGValues values) {
				super(Utils.getFilename(values.source.getLocation()));
				this.values = values;
				addActionListener(this);
			}

			@Override
			public void actionPerformed(ActionEvent e) {
				g_data.setHistogram(valuelist.indexOf(((HistItem) e.getSource()).values));
			}
		}

		public Menu() {
			super(new String[] { "File", "Edit", "Help", "Histograms" });

			getMenu(0).add(m_loadFile = new MenuItem("Load File"));
			getMenu(0).add(m_saveAll = new MenuItem("Save All Data"));
			getMenu(0).add(m_saveGraph = new MenuItem("Save Displayed Data"));
			getMenu(0).add(m_print = new MenuItem("Print Graph"));
			getMenu(0).add(m_export = new MenuItem("Export Graph"));
			getMenu(0).add(m_exit = new MenuItem("Exit"));

			getMenu(1).add(m_params = new MenuItem("Set Parameters"));
			getMenu(1).add(m_colorize = new MenuItem("Colorize"));
			getMenu(1).add(m_decolorize = new MenuItem("Decolorize"));
			getMenu(1).add(m_clear = new MenuItem("Clear Graphs"));

			getMenu(2).add(m_helpUsage = new MenuItem("Usage"));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			if (e.getSource() == m_loadFile) {
				g_fileDialog.setVisible(true, FileDialog.LOAD);
			} else if (e.getSource() == m_saveAll) {
				g_fileDialog.setVisible(true, FileDialog.SAVE);
			} else if (e.getSource() == m_saveGraph) {
				g_fileDialog.setVisible(true, FileDialog.SAVEG);
			} else if (e.getSource() == m_print) {
				g_data.print();
			} else if (e.getSource() == m_export) {
				g_exportDialog.setVisible(true);
			} else if (e.getSource() == m_exit) {
				Appl.stop();
			} else if (e.getSource() == m_params) {
				g_paramsDialog.setVisible(true);
			} else if (e.getSource() == m_colorize) {
				colormap.colorize();
				g_data.repaint();
			} else if (e.getSource() == m_decolorize) {
				colormap.decolorize();
				g_data.repaint();
			} else if (e.getSource() == m_clear) {
				valuelist.clear();
				clearHistograms();
				g_data.clear();
			} else if (e.getSource() == m_helpUsage) {
				g_helpDialog.setContent(Www.getPage("/tools/DistG/usage.hbi"));
			}
		}

		public void addHistogram(DistGValues values) {
			getMenu(3).add(new HistItem(values));
		}

		public void clearHistograms() {
			getMenu(3).removeAll();
		}
	}

	/**
	 * This is the dialog for loading and saving files
	 */
	class FileDialog extends FileSelectorDialog {
		int mode = 0;
		static final int LOAD = 1;
		static final int SAVE = 2;
		static final int SAVEG = 3;

		public FileDialog(String filename) {
			super("Select File", filename);
		}

		@Override
		public void afterOk(ActionEvent ev) {
			if (mode == LOAD) {
				new LoadFileExecutor(new Source(getFilename()), g_paramsDialog.getLenRange()).start();
			} else if (mode == SAVE) {
				new SaveDataExecutor(getFilename(), true).start();
			} else if (mode == SAVEG) {
				new SaveDataExecutor(getFilename(), false).start();
			}
		}

		public void setVisible(boolean state, int mode) {
			if (state) {
				switch (this.mode = mode) {
				case LOAD:
					setTitle("Load File");
					break;
				case SAVE:
					setTitle("Save Data to File");
					break;
				case SAVEG:
					setTitle("Save Data to File");
					break;
				}
			}
			super.setVisible(state);
		}
	}

	/**
	 * This is to export the graph to an image file
	 */
	class ExportDialog extends FileSelectorDialog {
		public ExportDialog() {
			super("Export graph", null);
		}

		@Override
		public void afterOk(ActionEvent ev) {
			g_data.export(getFilename());
		}
	}

	class ParamsFields extends GridBagPanel implements InputComponent {
		IntegerTextField g_minLen = null;
		IntegerTextField g_maxLen = null;
		IntegerTextField g_minCount = null;
		IntegerTextField g_maxCount = null;
		IntegerTextField g_bucketSize = null;
		IntegerTextField g_bucketCell = null;
		int row = 0;

		public ParamsFields() {
			super();
			createRow(this, "Min Seq Length", g_minLen = new IntegerTextField(params.lenRange.getMinVal()));
			createRow(this, "Max Seq Length", g_maxLen = new IntegerTextField(params.lenRange.getMaxVal()));
			createRow(this, "Min Count", g_minCount = new IntegerTextField(params.countRange.getMinVal()));
			createRow(this, "Max Count", g_maxCount = new IntegerTextField(params.countRange.getMaxVal()));
			createRow(this, "Bucket Size", g_bucketSize = new IntegerTextField(params.bucketSize));
			createRow(this, "Bucket Cell", g_bucketCell = new IntegerTextField(params.bucketCell));
		}

		public void createRow(GridBagPanel parent, String title, JComponent comp) {
			JLabel label = new JLabel(title);
			label.setPreferredSize(new Dimension(160, 20));
			parent.add(label, GridBagConstraints.HORIZONTAL, 0, row, 1, 0);
			comp.setPreferredSize(new Dimension(100, 20));
			parent.add(comp, GridBagConstraints.HORIZONTAL, 1, row++, 1, 0);
		}

		@Override
		public boolean isOk() {
			IntValueRange lenRange = g_paramsDialog.getLenRange();
			IntValueRange countRange = g_paramsDialog.getCountRange();
			boolean ok = true;
			ok = ok && Logger.info(lenRange.getMinVal() >= 0, "Seq length minimum too less!");
			ok = ok && Logger.info(lenRange.isValid(), "Seq length min/max problem!");
			ok = ok && Logger.info(countRange.getMinVal() >= 0, "Counter minimum too less!");
			ok = ok && Logger.info(countRange.isValid(), "Counter min/max problem!");
			ok = ok && Logger.info(g_bucketSize.getValue() > 0, "Bucket Size > 0!");
			return ok;
		}

		@Override
		public void reset() {
		}

		@Override
		public void saveState() {
		}
	}

	class ParamsDialog extends InputDialog<ParamsFields> {
		IntValueRange prevLenRange = null;
		int prevBucketCell = 0;

		public ParamsDialog() {
			super("Parameters", new ParamsFields(), 280, 170);
			prevLenRange = getLenRange();
		}

		@Override
		public void afterOk(ActionEvent ev) {
			params.lenRange = getLenRange();
			params.countRange = getCountRange();
			params.bucketSize = getBucketSize();
			params.bucketCell = getBucketCell();
			if (!prevLenRange.equals(params.lenRange) || prevBucketCell != params.bucketCell) {
				g_data.reload();
				prevLenRange = params.lenRange;
				prevBucketCell = params.bucketCell;
			} else {
				g_data.init();
			}
		}

		public IntValueRange getCountRange() {
			ParamsFields f = getInputComponent();
			return new IntValueRange(f.g_minCount.getValue(), f.g_maxCount.getValue());
		}

		public IntValueRange getLenRange() {
			ParamsFields f = getInputComponent();
			return new IntValueRange(f.g_minLen.getValue(), f.g_maxLen.getValue());
		}

		public int getBucketSize() {
			return getInputComponent().g_bucketSize.getValue();
		}

		public int getBucketCell() {
			return getInputComponent().g_bucketCell.getValue();
		}

		public void init(DistGParams params) {
			ParamsFields f = getInputComponent();
			f.g_minLen.setText(Integer.toString(params.lenRange.getMinVal()));
			f.g_maxLen.setText(Integer.toString(params.lenRange.getMaxVal()));
			f.g_minCount.setText(Integer.toString(params.countRange.getMinVal()));
			f.g_maxCount.setText(Integer.toString(params.countRange.getMaxVal()));
			f.g_bucketSize.setText(Integer.toString(params.bucketSize));
		}
	}

	class DataView extends JTabbedPane {
		DistGraph g_dist = null;
		HistGraph g_hist = null;

		public DataView() {
			setBorder(new Borders(5, 0, 0, 0));
			addTab("Distribution Graph", g_dist = new DistGraph(colormap));
			addTab("Histogram Graph", g_hist = new HistGraph());
		}

		public void add(int index) {
			try {
				g_dist.add(valuelist.get(index), params);
			} catch (Exception e) {
				Logger.error("Cannot add the file!", e);
			}
		}

		public void init() {
			try {
				g_dist.init(valuelist, params);
				if (valuelist.size() > 0) {
					g_hist.init(valuelist.get(0), params);
				} else {
					g_hist.clear();
				}
			} catch (Exception e) {
				Logger.error("Cannot generate!", e);
			}
		}

		public void reload() {
			for (int i = 0; i < valuelist.size(); i++) {
				Source source = valuelist.get(i).source;
				try {
					valuelist.set(i, DistGValues.load(source, params));
				} catch (Exception e) {
					Logger.error("Cannot reload " + source.getLocation(), e);
				}
			}
			init();
		}

		public void clear() {
			g_dist.clear();
			g_hist.clear();
		}

		public void setHistogram(int index) {
			try {
				g_hist.init(valuelist.get(index), params);
			} catch (Exception e) {
				Logger.error("Cannot set histogram!", e);
			}
		}

		public void print() {
			try {
				Print.toPrinter((GraphPanel) getSelectedComponent());
			} catch (Exception e) {
				Logger.error("Cannot print!", e);
			}
		}

		public void export(String filename) {
			try {
				Component comp = getSelectedComponent();
				if (comp instanceof CoordPanel) {
					((CoordPanel) comp).export(filename);
				}
			} catch (Exception e) {
				Logger.error("Cannot export!", e);
			}
		}

		public void saveValues(String filename, boolean fullrange) {
			try {
				DistGValues.save(valuelist, filename, new DistGParams(fullrange ? DistGValues.getTotalRange(valuelist)
						: params.lenRange, params.countRange, params.bucketSize, params.bucketCell));
			} catch (Exception e) {
				Logger.error(String.format("Cannot save file %s!", filename), e);
			}
		}
	}

	public static void main(String[] args) {
		try {
			new DistGAppl(args).run();
		} catch (Exception e) {
			Logger.error(BioinfException.convert(e).getMessage());
		}
	}

	public DistGAppl(String[] args) throws BioinfException {
		super("distg", args, true);
	}

	@Override
	public void run() throws BioinfException {
		colormap = new ColorMap();

		Logger.addListener(new Window("Distribution Graph", g_menu = new Menu(), g_data = new DataView()) {
			public String getCompId() {
				return "window";
			}
		});

		String filename = ApplParam.get("-f", Appl.cmdparams, (String) null);

		g_paramsDialog = new ParamsDialog();
		g_fileDialog = new FileDialog(filename);
		g_exportDialog = new ExportDialog();
		g_helpDialog = new HelpDialog("Help DistG");

		if (filename != null) {
			Source source = new Source(filename);
			if (!source.isUrl()) {
				g_fileDialog.setFilename(source.getLocation());
				g_fileDialog.setVisible(true, FileDialog.LOAD);
			} else {
				new LoadFileExecutor(source, g_paramsDialog.getLenRange()).start();
			}
		}
	}
}