package de.bioinf.appl.distg;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;

import de.bioinf.ui.ColorMap;
import de.bioinf.ui.ColorSelector;
import de.bioinf.ui.InputPopup;
import de.bioinf.ui.graph.CoordPanel;
import de.bioinf.ui.graph.GraphElem;
import de.bioinf.ui.graph.GraphPanel;
import de.bioinf.ui.graph.PolygonElem;
import de.bioinf.ui.graph.PolygonValuePair;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Utils;

public class DistGraph extends CoordPanel {
	private ColorMap colormap = null;
	private boolean initialized = false;

	private static final Color DEFAULT_COLOR = new Color(128, 128, 128);
	private static final double INITIAL_ZOOM = -0.02;

	private class DistColorSelectorPopup extends InputPopup<ColorSelector> implements MouseListener {
		private GraphElem elem = null;
		private GraphPanel panel = null;

		public DistColorSelectorPopup(GraphPanel panel) {
			super(panel, new ColorSelector(), 280, 120);
			this.panel = panel;
		}

		@Override
		public void afterOk(ActionEvent e) {
			colormap.put(elem.getId(), getInputComponent().getSelectedColor());
			panel.repaint();
		}

		public void mouseClicked(MouseEvent ev) {
			if (ev.getButton() == MouseEvent.BUTTON1) {
				GraphElem elem = (GraphElem) panel.find(ev.getX(), ev.getY());
				if (elem != null) {
					this.elem = elem;
					getInputComponent().setSelectedColor(elem.getBackground());
					setLocation(Math.min(ev.getX(), panel.getWidth() - getWidth()),
							Math.min(ev.getY(), panel.getHeight() - getHeight()));
					setVisible(true);
				}
			}
		}

		public void mouseEntered(MouseEvent e) {
		}

		public void mouseExited(MouseEvent e) {
		}

		public void mousePressed(MouseEvent e) {
		}

		public void mouseReleased(MouseEvent e) {
		}
	}

	private class DistElem extends PolygonElem {
		PolygonValuePair namepos = null;
		String name = null;

		public DistElem(PolygonValuePair[] values, String id) throws BioinfException {
			super(DistGraph.this, values, id);
			name = Utils.getFilename(id);
		}

		public Color getBackground() {
			Color c = colormap.get(getId());
			return c != null ? c : DEFAULT_COLOR;
		}

		public void paint(Graphics g) {
			super.paint(g);
			if (namepos != null) {
				g.setColor(getBackground());
				g.drawString(name, DistGraph.this.getX(namepos.xval), DistGraph.this.getY(namepos.yval));
			}
		}
	}

	private class DistMouseListener implements MouseListener {
		@Override
		public void mouseClicked(MouseEvent ev) {
			if (ev.getButton() != MouseEvent.BUTTON1) {
				DistElem elem = (DistElem) DistGraph.this.find(ev.getX(), ev.getY());
				if (elem != null) {
					elem.namepos = elem.namepos == null ? new PolygonValuePair(getXVal(ev.getX()), getYVal(ev.getY()))
							: null;
					repaint();
				}
			}
		}

		@Override
		public void mouseEntered(MouseEvent ev) {
		}

		@Override
		public void mouseExited(MouseEvent ev) {
		}

		@Override
		public void mousePressed(MouseEvent ev) {
		}

		@Override
		public void mouseReleased(MouseEvent ev) {
		}
	}

	/**
	 * Print an empty graph
	 */
	public DistGraph(ColorMap colormap) {
		super();

		this.colormap = colormap;
		setBackground(Color.WHITE);
		setInsets(new Insets(30, 50, 20, 5));

		addMouseListener(new DistColorSelectorPopup(this));
		addMouseListener(new DistMouseListener());
	}

	@Override
	public void clear() {
		super.clear();
		initialized = false;
		repaint();
	}

	/**
	 * Initialize the graph with the values and parameters.
	 */
	public void init(ArrayList<DistGValues> list, DistGParams params) throws BioinfException {
		super.clear();
		for (int i = 0; i < list.size(); i++) {
			_add(list.get(i), params);
		}
		zoom(INITIAL_ZOOM);
		initialized = true;
	}

	/**
	 * Add values to the graph
	 */
	public void add(DistGValues values, DistGParams params) throws BioinfException {
		_add(values, params);
		repaint();
	}

	@Override
	public void reset() {
		recalcRanges();
		zoom(INITIAL_ZOOM);
		repaint();
	}

	protected void postpaint(Graphics g) {
		super.postpaint(g);

		if (initialized) {
			// Write the header
			g.setColor(Color.BLACK);
			g.drawString("Distribution Graph", 5, 20);
			String time = Utils.getCurrentTime();
			g.drawString(time, getWidth() - g.getFontMetrics().stringWidth(time) - 5, 20);
			g.setColor(Color.LIGHT_GRAY);
			g.drawLine(5, 23, getWidth() - 5, 23);
		}
	}

	// **
	// Private
	// **

	private void _add(DistGValues values, DistGParams params) throws BioinfException {
		ArrayList<PolygonValuePair> pairs = new ArrayList<PolygonValuePair>();
		for (int i = params.countRange.getMinVal(); i <= params.countRange.getMaxVal(); i += params.bucketSize) {
			PolygonValuePair pair = new PolygonValuePair(i, 0);
			for (int c = 0; c < params.bucketSize && i + c < values.counts.length; c++) {
				pair.yval += values.counts[i + c];
			}
			pairs.add(pair);
		}
		add(new DistElem(pairs.toArray(new PolygonValuePair[pairs.size()]), values.source.getLocation()));
	}
}
