package de.bioinf.appl.distg;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import de.bioinf.ui.graph.CoordPanel;
import de.bioinf.ui.graph.GraphElem;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.Utils;
import de.bioinf.utils.ValueRange;

/**
 * The histogram graph is a collection of rectangles representing the counts for window of bucket size.
 * 
 * @author ng
 */
public class HistGraph extends CoordPanel {
	private String filename = null;
	private boolean initialized = false;

	private static final double INITIAL_ZOOM = -0.02;
	private static Color DEFAULT_COLOR = new Color(128, 128, 128);
	private static Color SELECTION_COLOR = new Color(192, 128, 128);

	private class HistMouseListener implements MouseListener {
		@Override
		public void mouseClicked(MouseEvent ev) {
			if (ev.getButton() != MouseEvent.BUTTON1) {
				HistElem elem = (HistElem) HistGraph.this.find(ev.getX(), ev.getY());
				if (elem != null) {
					elem.selected = !elem.selected;
					HistGraph.this.repaint();
				}
			}
		}

		@Override
		public void mouseEntered(MouseEvent ev) {
		}

		@Override
		public void mouseExited(MouseEvent ev) {
		}

		@Override
		public void mousePressed(MouseEvent ev) {
		}

		@Override
		public void mouseReleased(MouseEvent ev) {
		}
	}

	// Note: Add CORRECT to x value because the counting took place in the middle of a number.
	private class HistElem extends GraphElem {
		CoordPanel panel = null;
		int xval = 0;
		double yval = 0;
		int width = 0;
		boolean selected = false;

		private static final double CORRECT = 0.5;

		public HistElem(CoordPanel panel, int xval, double yval, int width, String id) throws BioinfException {
			super(id);
			this.panel = panel;
			this.xval = xval;
			this.yval = yval;
			this.width = width;
		}

		@Override
		public void adjustRanges(ValueRange xrange, ValueRange yrange) {
			xrange.add(xval);
			xrange.add(xval + width);
			yrange.add(yval);
		}

		@Override
		public boolean isVisible() {
			return panel.getX(xval) >= 0 && panel.getX(xval) <= panel.getWidth();
		}

		public void paint(Graphics g) {
			Color c = g.getColor();
			int x = panel.getX(xval - CORRECT);
			int y = panel.getY(0.0);
			int h = y - panel.getY(yval);
			g.setColor(selected ? SELECTION_COLOR : DEFAULT_COLOR);
			g.fillRect(x, y - h, panel.getX(xval + width - CORRECT) - x, h);
			g.setColor(Color.BLACK);
			g.drawRect(x, y - h, panel.getX(xval + width - CORRECT) - x, h);
			if (selected) {
				g.drawString(getId(), x, y - h - 1);
			}
			g.setColor(c);
		}

		@Override
		public boolean isHit(int x, int y) {
			return panel.getX(xval - CORRECT) < x && panel.getX(xval + width - CORRECT) > x;
		}
	}

	public HistGraph() {
		super();
		setBackground(Color.WHITE);
		setInsets(new Insets(30, 50, 20, 5));
		addMouseListener(new HistMouseListener());
	}

	public void clear() {
		super.clear();
		initialized = false;
		repaint();
	}

	public void init(DistGValues values, DistGParams params) throws BioinfException {
		super.clear();
		this.filename = Utils.getFilename(values.source.getLocation());
		for (int i = params.countRange.getMinVal(); i <= params.countRange.getMaxVal(); i += params.bucketSize) {
			double yval = 0;
			for (int c = 0; c < params.bucketSize && i + c < values.counts.length; c++) {
				yval += values.counts[i + c];
			}
			String id = Integer.toString(i)
					+ (params.bucketSize > 1 ? ".." + Integer.toString(i + params.bucketSize - 1) : "");
			add(new HistElem(this, i, yval, params.bucketSize, id));
		}
		zoom(INITIAL_ZOOM);
		initialized = true;
	}

	@Override
	public void reset() {
		recalcRanges();
		zoom(INITIAL_ZOOM);
		repaint();
	}

	protected void postpaint(Graphics g) {
		super.postpaint(g);

		if (initialized) {
			g.setColor(Color.BLACK);
			String time = Utils.getCurrentTime();
			g.drawString(time, getWidth() - g.getFontMetrics().stringWidth(time) - 5, 20);
			g.drawString(filename, 5, 20);
			g.setColor(Color.LIGHT_GRAY);
			g.drawLine(5, 23, getWidth() - 5, 23);
		}
	}
}
