package de.bioinf.base;

/**
 * A codon is a sequence of 3 nucleotids. The implementation depends on the DNA.
 * 
 * @author ng
 */
public class Codon {
	private String nucs = null;
	private int index = -1;

	/**
	 * All possible codons, mapping from index to codon.
	 */
	private static Codon[] codons = null;
	/**
	 * Number of possible codons (4^3).
	 */
	public static final int COUNT = 64;
	/**
	 * Length of a codon in number of nucleotids (that is constant)
	 */
	public static final int LENGTH = 3;

	private Codon(String nucs, int index) {
		super();
		this.nucs = nucs;
		this.index = index;
	}

	/**
	 * "Creates" a codon for the nucleotids. If any of the nucleotids is not
	 * valid or the length is <> 3 an "invalid codon" is returned.
	 */
	public static Codon create(String nucs) {
		if (nucs.length() != 3 || !Nucleotid.isValid(nucs)) {
			return new Codon(nucs, -1);
		}
		return codons[getIndex(nucs)];
	}

	public boolean equals(Object o) {
		if (o == null || !(o instanceof Codon)) {
			return false;
		}
		return index == ((Codon) o).index;
	}

	public static Codon getCodon(int index) {
		return codons[index];
	}

	/**
	 * Returns an (uniq) index for the codon. It must be true for a codon:
	 * Codon.getCodon(codon.getIndex))==codon!
	 */
	public int getIndex() {
		return index;
	}

	/**
	 * Returns the nucleotid at position pos - if unavailable it returns
	 * Nucleotid.UNKNOWN.
	 */
	public char getNucleotid(int pos) {
		return pos < nucs.length() ? nucs.charAt(pos) : Nucleotid.UNKNOWN;
	}

	public int hashCode() {
		return nucs.hashCode();
	}

	public boolean isValid() {
		return index >= 0;
	}

	/**
	 * Looks if the codon matches the pattern. The pattern is build of the
	 * nucleotids A,C,G,T and the wildcard ".", that stands for any nucleotid.
	 */
	public boolean matches(String pattern) {
		for (int i = 0; i < LENGTH; i++) {
			if ('.' != pattern.charAt(i) && nucs.charAt(i) != pattern.charAt(i)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Looks if the codon matches the patterns.
	 * 
	 * @see matches(String)
	 */
	public boolean matches(String[] patterns) {
		for (String pattern : patterns) {
			if (matches(pattern)) {
				return true;
			}
		}
		return false;
	}

	public String toString() {
		return nucs;
	}

	// **
	// Private
	// **

	/**
	 * Returns an index for the nucleotids. This index can be taken to point
	 * into the array codons.
	 */
	private static int getIndex(String nucs) {
		int index = 0;
		for (int i = 0; i < Codon.LENGTH; i++) {
			index = index * Nucleotid.COUNT + Nucleotid.getIndex(nucs.charAt(i));
		}
		return index;
	}

	static {
		codons = new Codon[COUNT];
		char[] nucs = new char[Codon.LENGTH];

		for (int i = 0; i < Nucleotid.COUNT; i++) {
			nucs[0] = Nucleotid.getDnaNucleotid(i);
			for (int j = 0; j < Nucleotid.COUNT; j++) {
				nucs[1] = Nucleotid.getDnaNucleotid(j);
				for (int k = 0; k < Nucleotid.COUNT; k++) {
					nucs[2] = Nucleotid.getDnaNucleotid(k);
					int index = (i * Nucleotid.COUNT + j) * Nucleotid.COUNT + k;
					codons[index] = new Codon(new String(nucs), index);
				}
			}
		}
	}
}
