package de.bioinf.base;

/**
 * A codon sequence is a nucleotid sequence collecting the nucleotis in groups
 * of three. If the nucleotid sequence is longer than dividable by 3, the
 * remaining nucleotids are not considered to be a codon, so they are not part
 * of this sequence!
 * 
 * @author ng
 */
public class CodonSequence extends NucleotidSequence {
	public CodonSequence(SequenceHeader info, String sequence) {
		super(info, sequence);
	}

	public CodonSequence(NucleotidSequence nseq) {
		this(nseq.header, nseq.sequence);
	}

	/**
	 * @see Codon.matches(pattern)
	 */
	public int count(String pattern) {
		int count = 0;
		for (int i = 0; i < length(); i++) {
			if (getCodon(i).matches(pattern)) {
				count++;
			}
		}
		return count;
	}

	/**
	 * @see Codon.matches(patterns)
	 */
	public int count(String[] patterns) {
		int count = 0;
		for (int i = 0; i < length(); i++) {
			if (getCodon(i).matches(patterns)) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Counts the occurances of the acids in the codon sequence.
	 */
	public int count(CodonMap map, AminoAcid[] acids) {
		int count = 0;
		for (int i = 0; i < length(); i++) {
			for (AminoAcid acid : acids) {
				if (acid.equals(map.getAcid(getCodon(i)))) {
					count++;
					break;
				}
			}
		}
		return count;
	}

	/**
	 * Returns the number of codons of this sequences. A codon with size < 3
	 * will NOT be added to the length.
	 */
	public int length() {
		return super.length() / 3;
	}

	/**
	 * Returns the index-th codon (or null).
	 */
	public Codon getCodon(int index) {
		if (index < length()) {
			return Codon.create(sequence.substring(index * 3, index * 3 + 3));
		} else {
			return null;
		}
	}
}
