package de.bioinf.ui;

import java.awt.Color;
import java.util.StringTokenizer;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.PersistentParams;

/**
 * A mapping from an id (String) to a color. A color map can be loaded and saved.
 * 
 * @see DefaultColorMap
 * 
 * @author ng
 */
public class ColorMap extends PersistentParams<String, Color> {
	protected static final Color DEFAULT_COLOR = new Color(128, 128, 128);

	/**
	 * Create an empty color map.
	 */
	public ColorMap() {
		super();
	}

	/**
	 * Create a color map from the specified file.
	 */
	public ColorMap(String filename) throws BioinfException {
		super(filename);
	}

	/**
	 * Copy constructor
	 */
	public ColorMap(ColorMap colormap) {
		super();
		for (String key : colormap.getKeys()) {
			put(key, colormap.getValue(key));
		}
	}

	/**
	 * Returns a new Sub-ColorMap for the specified ids.
	 */
	public ColorMap extract(String[] ids) {
		ColorMap colormap = new ColorMap();
		for (String id : ids) {
			colormap.put(id, isKey(id) ? getValue(id) : DEFAULT_COLOR);
		}
		return colormap;
	}

	/**
	 * Adds all not existing elements of map to the actual one.
	 */
	public void add(ColorMap colormap) {
		for (String key : colormap.getKeys()) {
			if (!isKey(key)) {
				put(key, colormap.getValue(key));
			}
		}
	}

	/**
	 * Inserts or replace all entries with the colormap
	 */
	public void put(ColorMap colormap) {
		for (String key : colormap.getKeys()) {
			put(key, colormap.getValue(key));
		}
	}

	/**
	 * Initialize with default color and return actual object.
	 */
	public ColorMap init(String[] ids) {
		clear();
		for (String id : ids) {
			put(id, DEFAULT_COLOR);
		}
		return this;
	}

	/**
	 * Assign generic colors to all elements of the map and return actual object
	 */
	public ColorMap colorize() {
		String[] ids = getKeys();
		int diff = (int) (160.0 / Math.pow(ids.length, 0.3333));
		int i = 0;
		for (int r = 64; r < 224; r += diff) {
			for (int g = 64; g < 224; g += diff) {
				for (int b = 64; b < 224; b += diff) {
					if (i < ids.length) {
						put(ids[i++], new Color(r, g, b));
					} else {
						return this;
					}
				}
			}
		}
		return this;
	}

	/**
	 * Remove all color information from this map (That means to set DEFAULT_COLOR for ell elements)
	 */
	public ColorMap decolorize() {
		String[] keys = getKeys();
		for (String key : keys) {
			put(key, DEFAULT_COLOR);
		}
		return this;
	}

	/**
	 * Returns the color for the id.
	 */
	public Color get(String id) {
		return getValue(id);
	}

	@Override
	public String getKeyObject(String key) {
		return key;
	}

	@Override
	public String[] getKeys() {
		return params.keySet().toArray(new String[params.size()]);
	}

	@Override
	public Color getValueObject(String s) throws BioinfException {
		StringTokenizer tok = new StringTokenizer(s, ",");
		try {
			int r = Integer.parseInt(tok.nextToken().trim());
			int g = Integer.parseInt(tok.nextToken().trim());
			int b = Integer.parseInt(tok.nextToken().trim());
			return new Color(r, g, b);
		} catch (Exception e) {
			throw new BioinfException(String.format("Invalid color '%s'", s));
		}
	}

	@Override
	public String toString(Color c) {
		return c.getRed() + "," + c.getGreen() + "," + c.getBlue();
	}
}
