package de.bioinf.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.border.EtchedBorder;

/**
 * A pane for typing or selecting input data. This dialog should be created once and enabled using the setVisible()
 * call. It is automatically disabled by the setVisible() call as well as activated.
 * 
 * @see InputDialog
 * @author ng
 */
public abstract class InputPopup<T extends JComponent> extends JPanel {
	private JComponent g_parent = null;
	private T g_input = null;

	class OkButton extends Button implements ActionListener {
		public OkButton() {
			super("Ok");
			addActionListener(this);
		}

		public void actionPerformed(ActionEvent e) {
			if (isOk()) {
				InputPopup.this.saveState();
				InputPopup.this.afterOk(e);
				InputPopup.this.setVisible(false);
			}
		}
	}

	class CancelButton extends Button implements ActionListener {
		public CancelButton() {
			super("Cancel");
			addActionListener(this);
		}

		public void actionPerformed(ActionEvent e) {
			InputPopup.this.reset();
			InputPopup.this.setVisible(false);
		}
	}

	/**
	 * Creates a dialog of width and height with the title.
	 * 
	 * @param parent - The parent component, used to reset the focus after disabling this popup with any button.
	 * @param comp - The component with the input fields.
	 */
	public InputPopup(JComponent parent, T comp, int width, int height) {
		super();

		setSize(width, height);
		setLayout(new BorderLayout());
		setBorder(new EtchedBorder());

		add(BorderLayout.CENTER, g_input = comp);
		JPanel buttons = new JPanel(new GridLayout(1, 2, 2, 0));
		buttons.add(new OkButton());
		buttons.add(new CancelButton());
		add(BorderLayout.SOUTH, buttons);

		g_parent = parent;
		g_parent.add(this);

		setVisible(false);
	}

	/**
	 * Implement this method to act on selecting the OK button and initial validation of the input fields.
	 * 
	 * @see isOk()
	 */
	abstract public void afterOk(ActionEvent e);

	/**
	 * Returns the component with the input fields.
	 */
	public T getInputComponent() {
		return g_input;
	}

	/**
	 * Check if the input fields are valid.
	 * 
	 * @see afterOk()
	 */
	public boolean isOk() {
		return isOk(g_input);
	}

	/**
	 * Reset to the previous state (either at initialization or after clicking on ok).
	 */
	public void reset() {
		reset(g_input);
	}

	/**
	 * Save the state to reset the component (after cancel click).
	 */
	public void saveState() {
		saveState(g_input);
	}

	@Override
	public void setVisible(boolean state) {
		super.setVisible(state);
		if (!state) {
			g_parent.requestFocus();
			g_parent.repaint();
		}
	}

	// **
	// Private
	// **

	private boolean isOk(Component comp) {
		if (comp instanceof InputComponent) {
			if (!((InputComponent) comp).isOk()) {
				return false;
			}
		}
		if (comp instanceof JComponent) {
			for (Component c : ((JComponent) comp).getComponents()) {
				if (!isOk(c)) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Reset all subcomponents
	 */
	private void reset(Component comp) {
		if (comp instanceof InputComponent) {
			((InputComponent) comp).reset();
		}
		if (comp instanceof JComponent) {
			for (Component c : ((JComponent) comp).getComponents()) {
				reset(c);
			}
		}
	}

	/**
	 * Save state of all subcomponents
	 */
	private void saveState(Component comp) {
		if (comp instanceof InputComponent) {
			((InputComponent) comp).saveState();
		}
		if (comp instanceof JComponent) {
			for (Component c : ((JComponent) comp).getComponents()) {
				saveState(c);
			}
		}
	}
}
