package de.bioinf.ui;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.bioinf.utils.Logger;
import de.bioinf.utils.Params;

/**
 * A partition is used to divide a set of values of type T into non overlapping subsets, each by a given partition name.
 * In general this is achieved as a hash map for values to partition names (partnames). Partitions may be empty.
 * 
 * External storing in a file may be done in the form
 * 
 * <pre>
 * part1=value11,value12,value13,...
 * part2=value21,value22,value23,...
 * </pre>
 * 
 * @author ng
 * 
 * @param <T>
 */
public abstract class Partition<T> implements Cloneable {
	private HashMap<T, String> map = new HashMap<T, String>();
	private HashSet<String> names = new HashSet<String>();

	public Partition() {
		super();
	}

	public Partition(HashMap<T, String> map) {
		this();
		Iterator<T> iter = map.keySet().iterator();
		while (iter.hasNext()) {
			T key = iter.next();
			String name = map.get(key);
			this.map.put(key, name);
			this.names.add(name);
		}
	}

	public Partition(Params<T, String> map) {
		this();
		T[] keys = map.getKeys();
		for (T key : keys) {
			String name = map.getValue(key);
			this.map.put(key, name);
			this.names.add(name);
		}
	}

	public Partition(T[] values) {
		this();
		for (T value : values) {
			map.put(value, null);
		}
	}

	/**
	 * Make a copy of this object.
	 */
	public Partition<T> copy() {
		Partition<T> p = null;
		try {
			p = (Partition<T>) super.clone();
			p.map = (HashMap<T, String>) map.clone();
			p.names = (HashSet<String>) names.clone();
		} catch (Exception e) {
			Logger.error(e);
		}
		return p;
	}

	/**
	 * Add (an empty) partition with the specified name.
	 */
	public void add(String pname) {
		names.add(pname);
	}

	/**
	 * Add a value to the "free" list - if not already existing.
	 */
	public void addValue(T value) {
		if (!map.containsKey(value)) {
			map.put(value, null);
		}
	}

	/**
	 * Add values to the "free" list - if not already existing.
	 */
	public void addValues(T[] values) {
		for (T value : values) {
			addValue(value);
		}
	}

	/**
	 * Link a value to an exisiting partition. Note: This automatically unlinks the value from another partition.
	 */
	public void link(T value, String pname) {
		if (names.contains(pname)) {
			map.put(value, pname);
		}
	}

	/**
	 * Unlink (free) a value from a partition
	 */
	public void unlink(T value) {
		map.put(value, null);
	}

	/**
	 * Is this partition already available? Even though it may not have elements.
	 */
	public boolean isPartition(String pname) {
		return names.contains(pname);
	}

	/**
	 * Returns the partition name for the value.
	 */
	public String getPartition(T value) {
		return map.get(value);
	}

	/**
	 * Remove the partition with the specified name. That means the values are free to participate in other partitions.
	 */
	public void deletePartition(String pname) {
		names.remove(pname);
		Set<T> keys = map.keySet();
		for (T key : keys) {
			if (pname.equals(map.get(key))) {
				map.put(key, null);
			}
		}
	}

	/**
	 * Returns all partition names, ordered.
	 */
	public String[] getPartitions() {
		String[] a = names.toArray(new String[names.size()]);
		Arrays.sort(a);
		return a;
	}

	public void renamePartition(String from, String to) {
		names.remove(from);
		names.add(to);
		Set<T> keys = map.keySet();
		for (T key : keys) {
			if (from.equals(map.get(key))) {
				map.put(key, to);
			}
		}
	}

	/**
	 * Returns all values.
	 */
	public T[] getValues() {
		return toArray(map.keySet());
	}

	/**
	 * Returns all values that have no partition name.
	 */
	public T[] getFreeValues() {
		HashSet<T> values = new HashSet<T>();
		for (T key : getValues()) {
			if (map.get(key) == null) {
				values.add(key);
			}
		}
		return toArray(values);
	}

	/**
	 * Returns all values of a specific partition.
	 */
	public T[] getValues(String name) {
		HashSet<T> values = new HashSet<T>();
		for (T key : getValues()) {
			if (name.equals(map.get(key))) {
				values.add(key);
			}
		}
		return toArray(values);
	}

	/**
	 * Initializes (frees) the partition
	 */
	public void init() {
		map.clear();
		names.clear();
	}

	/**
	 * Returns the partition as a string like
	 * 
	 * <pre>
	 * part1=va111,val12,val13,...
	 * part2=va121,val22,val33,...
	 * part3=
	 * part4=val4
	 * </pre>
	 * 
	 * Even empty partitions are printed!
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer();
		String[] pnames = getPartitions();
		for (String pname : pnames) {
			buf.append(pname + "=");
			T[] values = getValues(pname);
			for (int i = 0; i < values.length; i++) {
				if (i > 0) {
					buf.append(",");
				}
				buf.append(values[i]);
			}
			buf.append("\n");
		}
		return buf.toString();
	}

	/**
	 * Define this method which lets you specify the ordering of the elements in the values and partitions values
	 * section.
	 */
	abstract protected T[] toArray(Set<T> set);
}
