package de.bioinf.ui.graph;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.AbstractAction;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.ValueRange;

/**
 * This is a panel for coordinate displays. The (0,0) corner is (left,down), so the calculated value for the x position
 * (getX()) must take care of this!
 * 
 * @author ng
 */
public class CoordPanel extends GraphPanel {
	class ZoomInAction extends AbstractAction {
		public void actionPerformed(ActionEvent e) {
			CoordPanel.this.zoom(0.010101);
		}
	}

	class ZoomOutAction extends AbstractAction {
		public void actionPerformed(ActionEvent e) {
			CoordPanel.this.zoom(-0.01);
		}
	}

	class ResizeAction extends AbstractAction {
		public void actionPerformed(ActionEvent e) {
			CoordPanel.this.reset();
		}
	}

	class DeleteAction extends AbstractAction {
		public void actionPerformed(ActionEvent e) {
			GraphElem elem = CoordPanel.this.find(mousex, mousey);
			if (elem != null) {
				elem.setEnabled(false);
			}
			CoordPanel.this.repaint();
		}
	}

	/**
	 * This is a special MouseListener with the following options:
	 * <ul>
	 * <li>To click on an element the element's id is displayed (if exists).</li>
	 * <li>Pressing the CTRL-key and clicking somewhere activates selecting an area that will be taken as the next
	 * displayed area. Moving the mouse and releasing will finish the selection of the new visible area. Recalculation
	 * is done on xrange and yrange.</li>
	 * <li>Rotating the wheel leads to increasing or decreasing the size of the visible area (zooming)</li>
	 * </ul>
	 */
	class CoordMouseListener implements MouseListener, MouseMotionListener, MouseWheelListener {
		// The starting position of a selection (mouse pressed)
		private int x1, y1 = 0;
		// Is the rectangle/area selection active?
		private boolean selection = false;
		// The previous selected rectangle, just for undo the previous drawing.
		private Rectangle prerect = null;

		public CoordMouseListener() {
			super();
		}

		// Please note: change of min/max calls on the calculation in yrange!
		public void mouseReleased(MouseEvent e) {
			int x2 = e.getX();
			int y2 = e.getY();
			if (selection) {
				selection = false;
				prerect = null;
				xrange.set(getXVal(Math.min(x1, x2)), getXVal(Math.max(x1, x2)));
				yrange.set(getYVal(Math.max(y1, y2)), getYVal(Math.min(y1, y2)));
				repaint();
			}
		}

		public void mousePressed(MouseEvent e) {
			x1 = e.getX();
			y1 = e.getY();
			if (e.isControlDown()) {
				selection = true;
				prerect = null;
			}
		}

		public void mouseExited(MouseEvent e) {
		}

		public void mouseEntered(MouseEvent e) {
		}

		public void mouseClicked(MouseEvent e) {
			/*
			 * GraphElem elem = find(x1, y1); if (elem != null) { if (elem.getId() != null) { JComponent comp =
			 * (JComponent) e.getComponent(); Graphics g = comp.getGraphics(); int w =
			 * g.getFontMetrics().stringWidth(elem.getId()); int h = g.getFontMetrics().getHeight(); int x = x1 + w >
			 * comp.getWidth() ? x1 - w : x1; int y = y1 + h > comp.getHeight() ? y1 - h : y1;
			 * g.drawString(elem.getId(), x, y); } }
			 */
		}

		public void mouseDragged(MouseEvent e) {
			if (selection) {
				Graphics g = e.getComponent().getGraphics();
				g.setXORMode(Color.WHITE);
				if (prerect != null) {
					g.drawRect(prerect.x, prerect.y, prerect.width, prerect.height);
				}
				int x = (int) Math.min(x1, e.getX());
				int y = (int) Math.min(y1, e.getY());
				int w = (int) Math.abs(x1 - e.getX());
				int h = (int) Math.abs(y1 - e.getY());
				Rectangle rect = new Rectangle(x, y, w, h);
				g.drawRect(rect.x, rect.y, rect.width, rect.height);
				prerect = rect;
			} else {
				CoordPanel.this.move(x1 - e.getX(), e.getY() - y1);
				x1 = e.getX();
				y1 = e.getY();
			}
		}

		public void mouseMoved(MouseEvent e) {
		}

		public void mouseWheelMoved(MouseWheelEvent e) {
			zoom(-e.getWheelRotation() * 0.01);
		}
	}

	public CoordPanel() {
		this(new ValueRange(), new ValueRange());
	}

	public CoordPanel(ValueRange xrange, ValueRange yrange) {
		super(xrange, yrange);
		setInsets(new Insets(5, 50, 20, 5));

		CoordMouseListener listener = new CoordMouseListener();
		addMouseListener(listener);
		addMouseMotionListener(listener);
		addMouseWheelListener(listener);

		add(GraphKeyStroke.ZOOM_IN, new ZoomInAction());
		add(GraphKeyStroke.ZOOM_OUT, new ZoomOutAction());
		add(GraphKeyStroke.RESIZE, new ResizeAction());
		add(GraphKeyStroke.DELETE, new DeleteAction());
	}

	/**
	 * Returns the x position on the screen. This value is calculated from bottom up!
	 */
	public int getX(CoordElem elem) {
		return getX(elem.getXVal());
	}

	/**
	 * Returns the y position on the screen.
	 */
	public int getY(CoordElem elem) {
		return getY(elem.getYVal());
	}

	@Override
	protected void prepaint(Graphics g) {
		paintLines(g);
	}

	@Override
	protected void postpaint(Graphics g) {
	}

	@Override
	protected void recalc() {
	}

	/**
	 * Save the image into the file with the size of the displayed region.
	 */
	public void export(String filename) throws BioinfException {
		super.export(filename, getWidth(), getHeight() + insets.top + insets.bottom);
	}

	// **
	// Private
	// **

	/**
	 * Display the coordinate lines.
	 */
	private void paintLines(Graphics g) {
		if (elems.size() > 0) {
			double[] vals = yrange.getIntermediaryValues(5);
			for (int i = 0; i < vals.length; i++) {
				g.setColor(LINE_COLOR);
				int y = getYFraq((vals[i] - yrange.getMinVal()) / (yrange.getRange()));
				g.drawLine(0, y, getWidth() - insets.right, y);
				g.setColor(Color.BLACK);
				if ((int) vals[i] != vals[i]) {
					g.drawString(Double.toString(vals[i]), 0, y);
				} else {
					g.drawString(Integer.toString((int) vals[i]), 0, y);
				}
			}
			vals = xrange.getIntermediaryValues(5);
			for (int i = 0; i < vals.length; i++) {
				g.setColor(LINE_COLOR);
				int x = getXFraq((vals[i] - xrange.getMinVal()) / (xrange.getRange()));
				g.drawLine(x, 0 + insets.top, x, getHeight());
				g.setColor(Color.BLACK);
				if ((int) vals[i] != vals[i]) {
					g.drawString(Double.toString(vals[i]), x, getHeight());
				} else {
					g.drawString(Integer.toString((int) vals[i]), x, getHeight());
				}
			}
		}
	}
}
