package de.bioinf.ui.graph;

import java.awt.Color;
import java.awt.Graphics;

import javax.swing.JComponent;

import de.bioinf.ui.ColorMap;
import de.bioinf.utils.ValueRange;

/**
 * The base class for all elements that are displayed in a GraphPanel.
 * 
 * @author ng
 */
public abstract class GraphElem extends JComponent {
	// The id of the element
	private String id = null;
	// The mapping from an id (cmapid) to Color
	private ColorMap colormap = null;
	// The id pointing into the colormap
	private String cmapid = null;
	// Is the element disabled (out of sight)
	private boolean enabled = true;

	public GraphElem(String id) {
		super();
		this.id = id;
	}

	/**
	 * Returns the color of the element. Depending on colormap this will be used (or taken from getBackground()).
	 */
	public Color getBackground() {
		Color c = null;
		if (colormap != null) {
			c = colormap.get(cmapid);
		}
		return c != null ? c : super.getBackground();
	}

	/**
	 * Returns the id of the element.
	 */
	public String getId() {
		return id;
	}

	/**
	 * Is the element enabled?
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * Set a colormap to "override" the getBackground() color setting. The cmapid is the id pointing into the colormap.
	 */
	public void setColorMap(ColorMap map, String cmapid) {
		this.colormap = map;
		this.cmapid = cmapid;
	}

	/**
	 * Changes the enabled state. For interaction with the underlaying panel, see the subclasses of GraphPanel.
	 */
	public void setEnabled(boolean state) {
		enabled = state;
	}

	/**
	 * Adjust the ranges of the used x and y values (e.g. of the underlaying panel) by "adding" the ranges of the actual
	 * element.
	 */
	public abstract void adjustRanges(ValueRange xrange, ValueRange yrange);

	/**
	 * Is the Element hit on position?
	 */
	public abstract boolean isHit(int x, int y);

	/**
	 * Is the element seen/visible?
	 */
	public abstract boolean isVisible();

	/**
	 * Display the element.
	 */
	public abstract void paint(Graphics g);
}
