package de.bioinf.ui.graph;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.io.File;
import java.util.ArrayList;

import javax.imageio.ImageIO;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JPanel;

import de.bioinf.appl.Environment;
import de.bioinf.utils.BioinfException;
import de.bioinf.utils.ValueRange;

/**
 * This panel is a specialized panel without a layout manager to paint graphics. Graphics means coordinate elements,
 * functions and other things that are printed for itself.
 * 
 * The graph panel automatically sets the mouse positions mouesex, mousey on mouse movement.
 * 
 * @author ng
 */
public abstract class GraphPanel extends JPanel implements Printable, Cloneable {
	// For printing (on a printer!)
	public static final int DPI_72 = 1;
	public static final int DPI_144 = 2;
	public static final int DPI_288 = 4;

	public static final Color LINE_COLOR = Color.LIGHT_GRAY;
	public static final Color BORDER_COLOR = new Color(230, 230, 230);

	// The mouse position
	protected int mousex = -1;
	protected int mousey = -1;

	/**
	 * The value range of all x coordinate values.
	 */
	protected ValueRange xrange = null;
	/**
	 * The value range of all y coordinate values.
	 */
	protected ValueRange yrange = null;
	/**
	 * The space between the view of the graphical components and the borders of the panel. This space may be used for
	 * information texts, or for the texts of the coordinate lines, that are not overwritten by the graphical elements.
	 */
	protected Insets insets = new Insets(0, 20, 20, 0);
	/**
	 * The elements displayed in this graph.
	 */
	protected ArrayList<GraphElem> elems = new ArrayList<GraphElem>();

	/**
	 * Has some of the elements a change? That may lead to a recalc() of the internal values.
	 */
	private boolean changed = false;

	/**
	 * A listener to set mousex and mousey.
	 */
	class GraphMouseListener implements MouseMotionListener {
		public GraphMouseListener() {
			super();
		}

		public void mouseMoved(MouseEvent e) {
			mousex = e.getX();
			mousey = e.getY();
		}

		public void mouseDragged(MouseEvent e) {
		}
	}

	public GraphPanel() {
		this(new ValueRange(), new ValueRange());
	}

	public GraphPanel(ValueRange xrange, ValueRange yrange) {
		super(null);
		this.xrange = xrange;
		this.yrange = yrange;
		this.setFocusable(true);
		setBackground(Color.WHITE);
		addMouseMotionListener(new GraphMouseListener());
	}

	/**
	 * Adds a key stroke. To react on this stroke, their corresponding action object (e.g. AbstractAction) must be
	 * implemented in a subclass.
	 */
	public void add(GraphKeyStroke stroke, Action action) {
		getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(stroke.getStroke(), stroke.getStroke());
		getActionMap().put(stroke.getStroke(), action);
	}

	/**
	 * Adds a graphical element to the panel.
	 */
	public void add(GraphElem elem) {
		elems.add(elem);
		elem.adjustRanges(xrange, yrange);
	}

	/**
	 * Returns (all - the deactived also!) graphic elements.
	 */
	public ArrayList<GraphElem> getGraphElems() {
		return elems;
	}

	/**
	 * Returns the index-th element.
	 */
	public GraphElem getGraphElem(int index) {
		return elems.get(index);
	}

	/**
	 * Find a graph element at position (x,y). If not found return null. The order of adding is important, so the last
	 * added will be returned, because it is printed on the top!
	 */
	public GraphElem find(int x, int y) {
		for (int i = elems.size() - 1; i >= 0; i--) {
			GraphElem elem = elems.get(i);
			if (elem.isEnabled() && elem.isHit(x, y)) {
				return elem;
			}
		}
		return null;
	}

	/**
	 * Finds and returns all elements at the position (x,y). If none found return empty list.
	 */
	protected ArrayList<GraphElem> findAll(int x, int y) {
		ArrayList<GraphElem> a = new ArrayList<GraphElem>();
		for (int i = elems.size() - 1; i >= 0; i--) {
			GraphElem elem = elems.get(i);
			if (elem.isEnabled() && elem.isHit(x, y)) {
				a.add(elem);
			}
		}
		return a;
	}

	public double getMinXVal() {
		return xrange.getMinVal();
	}

	public double getMaxXVal() {
		return xrange.getMaxVal();
	}

	public double getMinYVal() {
		return yrange.getMinVal();
	}

	public double getMaxYVal() {
		return yrange.getMaxVal();
	}

	public ValueRange getXRange() {
		return xrange;
	}

	public ValueRange getYRange() {
		return yrange;
	}

	/**
	 * Returns the x position on the screen.
	 */
	public int getX(double xval) {
		return getXFraq((xval - xrange.getMinVal()) / (xrange.getRange()));
	}

	/**
	 * Returns the x position on the screen. This value is calculated from bottom up!
	 * 
	 * @param frag - The fragment part 0..1
	 */
	public int getXFraq(double frag) {
		return (int) (frag * (double) (getWidth() - (insets.left + insets.right)) + insets.left);
	}

	/**
	 * Returns the x value for the x position
	 */
	public double getXVal(int x) {
		return (double) (x - insets.left) / (double) (getWidth() - (insets.left + insets.right)) * xrange.getRange()
				+ xrange.getMinVal();
	}

	/**
	 * Returns the y position on the screen.
	 */
	public int getY(double yval) {
		return getYFraq((yval - yrange.getMinVal()) / (yrange.getRange()));
	}

	/**
	 * Returns the y position on the screen.
	 * 
	 * @param frag - The fragment part 0..1
	 */
	public int getYFraq(double frag) {
		return (int) ((1 - frag) * (double) (getHeight() - (insets.top + insets.bottom)) + insets.top);
	}

	/**
	 * Returns the y value for the y position. Notice, that the position 0 is on the bottom of the panel!
	 */
	public double getYVal(int y) {
		return (double) (getHeight() - y - insets.bottom) / (double) (getHeight() - (insets.top + insets.bottom))
				* yrange.getRange() + yrange.getMinVal();
	}

	/**
	 * Removes all graph elements and initializes the ranges.
	 */
	public void clear() {
		elems.clear();
		xrange.init();
		yrange.init();
	}

	/**
	 * Calculates the xrange and yrange anew, regarding only active/enabled elements.
	 */
	public void recalcRanges() {
		xrange.init();
		yrange.init();
		for (GraphElem elem : elems) {
			if (elem.isEnabled()) {
				elem.adjustRanges(xrange, yrange);
			}
		}
	}

	/**
	 * Informs the panel, that some elements (or the graph itself) may have changed (e.g. height, visibility,...). It is
	 * "used" at the next paint() call, doing a recalc().
	 */
	public void setChanged() {
		changed = true;
	}

	/**
	 * Paint is done in following steps:
	 * <ul>
	 * <li>The border lines of the visible area for "the" graphical elements are drawn. The visible area is the panel's
	 * area "subtracting" the inset values.</li>
	 * <li>prepaint() is called. (e.g. In general the coordinate lines are painted.)</li>
	 * <li>The graphical elements are printed inside the visible area.</li>
	 * <li>postpaint() is called. That means to be able to print common information inside the graph's panel.</li>
	 * </ul>
	 */
	public void paint(Graphics g) {
		super.paintComponent(g);

		if (changed) {
			recalc();
			changed = false;
		}

		g.setColor(BORDER_COLOR);
		Rectangle rect = getInnerClip();
		g.drawRect(rect.x, rect.y, rect.width, rect.height);

		prepaint(g);

		// Paint elements
		Shape clip = g.getClip();
		g.setClip(rect);
		g.setFont(Environment.DEFAULT_FONT);
		for (GraphElem elem : elems) {
			if (elem.isEnabled() && elem.isVisible()) {
				elem.paint(g);
			}
		}
		g.setClip(clip);

		postpaint(g);

		super.paintChildren(g);
	}

	/**
	 * A method to paint something before the "real" elements are printed, so the elements are not overdrawn. For
	 * example print some "coordinate"-lines (in a coordinate system) before the elements are displayed.
	 */
	protected abstract void prepaint(Graphics g);

	/**
	 * A method to paint something after the "real" elements are printed.
	 */
	protected abstract void postpaint(Graphics g);

	/**
	 * Recalculate internal values (may be used on a change (see setChange())
	 */
	protected abstract void recalc();

	/**
	 * Prepares the actual object to be printed to a printer (or file). That means to print into a new graphics
	 * <code>g</code> the elements with the size of the paper (see <code>PageFormat</code>).
	 */
	public int print(Graphics g, PageFormat pf, int pageno) throws PrinterException {
		int dpiindex = DPI_144;

		if (pageno != 0) {
			return NO_SUCH_PAGE;
		}

		Dimension size = getSize();

		int width = (int) Math.round(pf.getImageableWidth() * dpiindex) - 3;
		int height = (int) Math.round(pf.getImageableHeight() * dpiindex) - 3;
		setSize(width, height);

		try {
			Graphics2D g2d = (Graphics2D) g;
			g2d.translate(pf.getImageableX(), pf.getImageableY());
			g2d.scale(1.0 / dpiindex, 1.0 / dpiindex);
			paint(g2d);
		} catch (Exception e) {
			throw new PrinterException(e.getMessage());
		}

		setSize(size);

		return PAGE_EXISTS;
	}

	/**
	 * Save the actual graph into a png file called filename with specified size.
	 */
	public void export(String filename, int width, int height) throws BioinfException {
		try {
			Dimension size = getSize();

			setSize(width, height);
			BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
			paint(image.createGraphics());
			ImageIO.write(image, "png", new File(filename));

			setSize(size);
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(String.format("Cannot write to %s!", filename));
		}
	}

	/**
	 * Move the viewable area by dx/dy pixels.
	 */
	public void move(int dx, int dy) {
		if (dx != 0 || dy != 0) {
			double delta = ((double) dx / (getWidth() - (insets.left + insets.right))) * xrange.getRange();
			xrange.set(xrange.getMinVal() + delta, xrange.getMaxVal() + delta);
			delta = ((double) dy / (getHeight() - (insets.top + insets.bottom))) * yrange.getRange();
			yrange.set(yrange.getMinVal() + delta, yrange.getMaxVal() + delta);
			repaint();
		}
	}

	/**
	 * Recalculates all ranges and repaint().
	 */
	public void reset() {
		recalcRanges();
		repaint();
	}

	/**
	 * Zoom the visible area and repaint the content.
	 */
	public void zoom(double factor) {
		double delta = factor * xrange.getRange();
		xrange.set(xrange.getMinVal() + delta, xrange.getMaxVal() - delta);
		delta = factor * yrange.getRange();
		yrange.set(yrange.getMinVal() + delta, yrange.getMaxVal() - delta);
		repaint();
	}

	/**
	 * Set the space between the graphical components and the borders of the panel. This space may be used for
	 * information texts, or for the texts of the coordinate lines, that are not overwritten by the graphical elements.
	 */
	public void setInsets(Insets insets) {
		this.insets = insets;
	}

	/**
	 * Get the clipping area inside the insets.
	 */
	public Rectangle getInnerClip() {
		return new Rectangle(insets.left, insets.top, getWidth() - (insets.right + insets.left), getHeight()
				- (insets.bottom + insets.top));
	}
}
