package de.bioinf.ui.graph;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JScrollBar;
import javax.swing.UIManager;

import de.bioinf.utils.BioinfException;
import de.bioinf.utils.ValueRange;

// TODO staticXRange: is there a better solution for setting the value and overwrite already made calculations?
/**
 * This is a panel for horizontal bar displays. That means the added elements are lines that represent value ranges of a
 * specific object/data.
 * 
 * @author ng
 */
public class HorizBarPanel extends GraphPanel {
	/**
	 * The height of all lines together
	 */
	private int totalHeight = 0;
	/**
	 * A fixed range for the displayed x values
	 */
	private ValueRange staticXRange = null;
	/**
	 * A scrollbar on the right side
	 */
	private BarScrollBar g_scrollBar = null;
	private YOffset yoffset = new YOffset();

	// The space between the lines.
	private static final int LINE_SPACE = 2;
	private static final int SCROLLBAR_WIDTH = UIManager.getInt("ScrollBar.width");
	private static final int SCROLLBAR_GAP = 3;

	/**
	 * The offset of the display. This leads to moving the display up or down.
	 */
	class YOffset {
		private int offset = 0;

		public void set(int offset, boolean scrollbarAdjust) {
			if (offset + totalHeight + (insets.top + insets.bottom) < getHeight()) {
				offset = getHeight() - (totalHeight + (insets.top + insets.bottom));
			}
			offset = Math.min(offset, 0);
			if (this.offset != offset) {
				this.offset = offset;
				if (scrollbarAdjust) {
					g_scrollBar.adjust();
				}
				setChanged();
				repaint();
			}
		}

		public void add(int diff, boolean scrollbarAdjust) {
			set(offset + diff, scrollbarAdjust);
		}
	}

	class BarMouseListener implements MouseListener, MouseMotionListener, MouseWheelListener {
		private int y1 = 0;

		public BarMouseListener() {
			super();
		}

		public void mouseReleased(MouseEvent e) {
		}

		public void mousePressed(MouseEvent e) {
			y1 = e.getY();
		}

		public void mouseExited(MouseEvent e) {
		}

		public void mouseEntered(MouseEvent e) {
		}

		public void mouseClicked(MouseEvent e) {
		}

		public void mouseDragged(MouseEvent e) {
			yoffset.add(-(y1 - e.getY()), true);
			y1 = e.getY();
		}

		public void mouseMoved(MouseEvent e) {
		}

		public void mouseWheelMoved(MouseWheelEvent e) {
			for (GraphElem elem : elems) {
				if (((HorizBarElem) elem).getY() >= 0) {
					yoffset.add(-e.getWheelRotation() * (elem.getHeight() + LINE_SPACE), true);
					return;
				}
			}
		}
	}

	class BarComponentListener implements ComponentListener {
		public void componentHidden(ComponentEvent e) {
		}

		public void componentMoved(ComponentEvent e) {
		}

		public void componentResized(ComponentEvent e) {
			// setYOffset(yoffset);
			g_scrollBar.resize();
		}

		public void componentShown(ComponentEvent e) {
		}
	}

	class BarScrollBar extends JScrollBar implements AdjustmentListener {
		boolean active = true;

		public BarScrollBar() {
			super(JScrollBar.VERTICAL);
			addAdjustmentListener(this);
		}

		// TODO resize only if visible!
		public void resize() {
			HorizBarPanel p = HorizBarPanel.this;
			setLocation(p.getWidth() - p.insets.right + SCROLLBAR_GAP, p.insets.top);
			setSize(SCROLLBAR_WIDTH, p.getHeight() - (p.insets.top + p.insets.bottom));
			setVisibleAmount(p.getHeight() - (p.insets.top + p.insets.bottom));
			setMinimum(0);
			setMaximum(totalHeight);
		}

		public void adjust() {
			active = false;
			setValue(-yoffset.offset);
			active = true;
		}

		@Override
		public void adjustmentValueChanged(AdjustmentEvent ev) {
			if (active) {
				if (ev.getAdjustmentType() == AdjustmentEvent.TRACK) {
					yoffset.set(-ev.getValue(), false);
				}
			}
		}
	}

	public HorizBarPanel() {
		this(new ValueRange());
	}

	public HorizBarPanel(ValueRange xrange) {
		super(xrange, new ValueRange());
		BarMouseListener listener = new BarMouseListener();
		addMouseListener(listener);
		addMouseMotionListener(listener);
		addMouseWheelListener(listener);
		addComponentListener(new BarComponentListener());
		add(g_scrollBar = new BarScrollBar());
	}

	/**
	 * Adds an element to this panel.
	 */
	public void add(HorizBarElem elem) {
		super.add(elem);
		setChanged();
	}

	/**
	 * Returns the minimum (left) x position of the element on the screen.
	 */
	public int getMinX(HorizBarElem elem) {
		return getXFraq((elem.getMinXVal() - xrange.getMinVal()) / (xrange.getRange()));
	}

	/**
	 * Returns the maximum (right) x position of the element on the screen.
	 */
	public int getMaxX(HorizBarElem elem) {
		return getXFraq((elem.getMaxXVal() - xrange.getMinVal()) / (xrange.getRange()));
	}

	/**
	 * Set the y offset.
	 */
	public void setYOffset(int offset) {
		yoffset.set(offset, true);
	}

	/**
	 * Removes all graphical elements from this graph.
	 */
	public void clear() {
		super.clear();
		setYOffset(0);
	}

	/**
	 * Save the graph into the file with the total height of all elements.
	 * 
	 * Note: It is a bit tricky to do, because you have to make sure that the printout will start on the top and have
	 * the total height of all elements, export then and reset to actual size. This can not been done in a thread,
	 * otherwise the graphical unit will in between act on the changed parameters and reset to the best fit.
	 */
	public void export(String filename) throws BioinfException {
		setScrollBarVisible(false);
		int offset = yoffset.offset;
		yoffset.set(0, false);
		recalc();
		super.export(filename, getWidth(), totalHeight + insets.top + insets.bottom);
		setScrollBarVisible(true);
		yoffset.set(offset, false);
		recalc();
	}

	@Override
	public void setInsets(Insets insets) {
		insets.right += SCROLLBAR_WIDTH + SCROLLBAR_GAP;
		super.setInsets(insets);
	}

	public void setScrollBarVisible(boolean state) {
		if (state) {
			g_scrollBar.setVisible(true);
			insets.right += SCROLLBAR_WIDTH + SCROLLBAR_GAP;
		} else {
			g_scrollBar.setVisible(false);
			insets.right -= SCROLLBAR_WIDTH + SCROLLBAR_GAP;
		}
	}

	/**
	 * Set an x coordinate range that will be always the default range to display. Passing null or an invalid range as
	 * parameter disables this settings.
	 * 
	 * @see recalc()
	 */
	public void setStaticXRange(ValueRange range) {
		staticXRange = range != null && range.isValid() ? range : null;
		setChanged();
		repaint();
	}

	@Override
	protected void prepaint(Graphics g) {
		paintLines(g);
	}

	@Override
	protected void postpaint(Graphics g) {
	}

	/**
	 * Calculate the total height anew. This must be done e.g. if an element's height changes, or an element is removed.
	 */
	protected void recalc() {
		recalcRanges();
		if (staticXRange != null) {
			xrange.set(staticXRange.getMinVal(), staticXRange.getMaxVal());
		}
		totalHeight = 0;
		for (GraphElem ge : elems) {
			HorizBarElem elem = (HorizBarElem) ge;
			if (elem.isEnabled()) {
				elem.setY(totalHeight + yoffset.offset + insets.top);
				totalHeight += elem.getHeight() + LINE_SPACE;
			}
		}
		g_scrollBar.resize();
	}

	// **
	// Private
	// **

	/**
	 * Display the vertical lines for the value ranges.
	 */
	private void paintLines(Graphics g) {
		if (elems.size() > 0) {
			double[] vals = xrange.getIntermediaryValues(5);
			for (int i = 0; i < vals.length; i++) {
				g.setColor(LINE_COLOR);
				int x = getXFraq((vals[i] - xrange.getMinVal()) / (xrange.getRange()));
				g.drawLine(x, insets.top, x, getHeight());
				g.setColor(Color.BLACK);
				if ((int) vals[i] != vals[i]) {
					g.drawString(Double.toString(vals[i]), x, getHeight());
				} else {
					g.drawString(Integer.toString((int) vals[i]), x, getHeight());
				}
			}
		}
	}
}
