package de.bioinf.ui.graph;

import java.awt.Color;
import java.awt.Graphics;

import de.bioinf.utils.ValueRange;

/**
 * This spline component consists of n y-values each of them in a fixed distance. For more information about he used
 * Catmull spline algorithm see:
 * 
 * <a href="http://stackoverflow.com/questions/939874/is-there-a-java-library-with-3d-spline-functions">here</a>
 * 
 * and
 * 
 * <a href="http://www.mvps.org/directx/articles/catmull/">here</a>
 * 
 * @author ng
 */
public class SplineElem extends GraphElem {
	private CoordPanel panel = null;
	private double[] yvalues = null;
	private ValueRange range = null;
	private int points = 0;

	// The size of the displayed area
	private ValueRange xrangeElem = new ValueRange();
	private ValueRange yrangeElem = new ValueRange();

	/**
	 * Creates a spline element for n y-values of the same distance on the x axis.
	 * 
	 * @param panel - The underlaying graphics panel
	 * @param range - The range for the x values
	 * @param yvalues - The y values in the specified range (all with the same distance)
	 * @param points - The number of points between each y value pair specified
	 * @param id - The id of this elements.
	 */
	public SplineElem(CoordPanel panel, ValueRange range, double[] yvalues, int points, String id) {
		super(id);

		this.panel = panel;
		this.range = range;
		this.points = points;

		// Copy and add 2 end points
		this.yvalues = new double[yvalues.length + 2];
		System.arraycopy(yvalues, 0, this.yvalues, 1, yvalues.length);
		this.yvalues[0] = yvalues[0];
		this.yvalues[this.yvalues.length - 1] = yvalues[yvalues.length - 1];
	}

	@Override
	public void adjustRanges(ValueRange xrange, ValueRange yrange) {
		xrange.add(range);
		xrangeElem.add(range);
		for (double yval : yvalues) {
			yrange.add(yval);
			yrangeElem.add(yval);
		}
	}

	@Override
	public boolean isHit(int x, int y) {
		return false;
	}

	@Override
	public boolean isVisible() {
		return panel.getX(xrangeElem.getMaxVal()) >= 0 && panel.getX(xrangeElem.getMinVal()) <= panel.getWidth()
				&& panel.getY(yrangeElem.getMinVal()) > 0 && panel.getY(yrangeElem.getMaxVal()) <= panel.getHeight();
	}

	@Override
	public void paint(Graphics g) {
		Color c = g.getColor();
		g.setColor(Color.BLACK);

		double dx = (double) range.getRange() / (double) ((yvalues.length - 3) * (points + 1));
		double dt = 1.0 / (points + 1);
		double x1 = range.getMinVal();
		double y1 = yvalues[1];

		for (int i = 0; i < yvalues.length - 3; i++) {
			// g.drawOval(panel.getX(x1) - 3, panel.getY(y1) - 2, 7, 7);
			double x2 = x1;
			double t = dt;

			for (int k = 0; k <= points; k++, t += dt) {
				double y2 = calc(t, yvalues[i], yvalues[i + 1], yvalues[i + 2], yvalues[i + 3]);
				g.drawLine(panel.getX(x1), panel.getY(y1), panel.getX(x2 += dx), panel.getY(y2));
				x1 = x2;
				y1 = y2;
			}
		}

		g.setColor(c);
	}

	// **
	// Private
	// **

	// Calculate intermediate points between p1..p2, when t is 0..1
	private double calc(double t, double p0, double p1, double p2, double p3) {
		return 0.5 * ((2 * p1) + (-p0 + p2) * t + (2 * p0 - 5 * p1 + 4 * p2 - p3) * (t * t) + (-p0 + 3 * p1 - 3 * p2 + p3)
				* (t * t * t));
	}
}
