package de.bioinf.ui.graph;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Insets;
import java.util.Arrays;

import de.bioinf.utils.Logger;
import de.bioinf.utils.Utils;

/**
 * This class displays a windowed "envelope" curve of a set of values as a
 * graph. On the x axis the values itself are measured on the y axis a
 * "statistical" amount is shown that shows the number of values in a window.
 * This window is moving through the whole value area.
 * 
 * @author ng
 */
public class WindowGraph extends CoordPanel {
	private Double[] values = null;
	private String info = null;
	// Window count
	private int wcount = 0;
	// Window width
	private double wwidth = 0.0;

	public WindowGraph() {
		super();
		setBackground(Color.WHITE);
		setInsets(new Insets(30, 50, 20, 5));
	}

	/**
	 * (Re)display the graph.
	 * 
	 * @param v - The array of the values
	 * @param wwidth - The width of the windows / frames.
	 */
	public void display(Double[] v, int wcount, double wwidth) {
		Logger.info("Preparing values for display ...");

		clear();
		this.wcount = wcount;
		this.wwidth = wwidth;

		Double[] ds = v.clone();
		Arrays.sort(ds);
		this.values = ds;

		double min = ds[0];
		double max = ds[ds.length - 1];
		double dx = (max - min - wwidth) / wcount;
		double x = min;

		PolygonValuePair[] pairs = new PolygonValuePair[wcount];
		for (int i = 0; i < pairs.length; i++) {
			int count = 0;
			for (int k = 0; k < ds.length && ds[k] < x + wwidth; k++) {
				if (x <= ds[k]) {
					count++;
				}
			}
			pairs[i] = new PolygonValuePair();
			pairs[i].xval = x + (wwidth / 2.0);
			pairs[i].yval = (count / wwidth) / (v.length / (max - min));
			x += dx;
		}
		try {
			add(new PolygonElem(this, pairs, null));
		} catch (Exception e) {
		}

		repaint();
		Logger.info("");
	}

	/**
	 * Returns the calculation window width
	 */
	public double getWWidth() {
		return wwidth;
	}

	public Double[] getValues() {
		return values;
	}

	public void setInfo(String info) {
		this.info = info;
	}

	@Override
	protected void prepaint(Graphics g) {
		super.prepaint(g);

		if (xrange.isValid()) {
			Color c = g.getColor();

			g.setColor(Color.BLACK);
			double minval = xrange.getMinVal();
			g.drawLine(getX(minval) + 5, insets.top + 5, getX(minval + wwidth) + 5, insets.top + 5);
			g.drawLine(getX(minval) + 5, insets.top + 3, getX(minval) + 5, insets.top + 7);
			g.drawLine(getX(minval + wwidth) + 5, insets.top + 3, getX(minval + wwidth) + 5, insets.top + 7);

			g.setColor(c);
		}
	}

	@Override
	protected void postpaint(Graphics g) {
		super.postpaint(g);
		Color c = g.getColor();

		// Write the header
		g.setColor(Color.BLACK);
		g.drawString(String.format("%s [wc=%d, ws=%f]", info != null ? info : "?", wcount, wwidth), 5, 20);
		String time = Utils.getCurrentTime();
		g.drawString(time, getWidth() - g.getFontMetrics().stringWidth(time) - 5, 20);
		g.setColor(Color.LIGHT_GRAY);
		g.drawLine(5, 23, getWidth() - 5, 23);

		g.setColor(c);
	}
}
