package de.bioinf.utils;

import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * This class is used to extract special parts of strings using a pattern. The special parts to
 * extract are marked by "#". The pattern consists of following elements:
 * 
 * <ul>
 * <li> "*": a character sequence (length 0..n)
 * <li> ".": a character (length exactly 1!)
 * <li> "#": a character sequence that should be extracted (up to the next character in the pattern)
 * <li> "\": the following character is taken as is (escaped)
 * <li> else: the character must be found in the analyzed string.
 * </ul>
 * 
 * <p>
 * Remark: The special characters '*' and '#' in the pattern should not be used in combination
 * directly together, e.g. "**" or "#*", because the patterns cannot be correct evaluated (where is
 * the ending position?).
 * 
 * @author ng
 */
public class Extractor {
	private String pattern = null;

	public Extractor(String pattern) {
		super();
		this.pattern = pattern;
	}

	/**
	 * Returns the parts (described by "#" in pattern) as an array of strings.
	 */
	public ArrayList<String> get(String str) {
		ArrayList<String> parts = new ArrayList<String>();

		int pos = 0;
		boolean escaped = false;
		boolean consume = false;
		StringTokenizer tok = new StringTokenizer(pattern, "#*.\\", true);
		while (tok.hasMoreTokens() && pos >= 0) {
			String token = tok.nextToken();
			if (!escaped) {
				if ("#".equals(token)) {
					consume = true;
					if (!tok.hasMoreTokens()) {
						parts.add(str.substring(pos));
					}
					continue;
				} else if ("*".equals(token)) {
					continue;
				} else if (".".equals(token)) {
					pos++;
					continue;
				} else if ("\\".equals(token)) {
					escaped = true;
					continue;
				}
			} else {
				escaped = false;
			}
			int lastpos = pos;
			pos = str.indexOf(token, pos);
			if (pos >= 0) {
				if (consume) {
					parts.add(str.substring(lastpos, pos));
					consume = false;
				}
				pos += token.length();
			}
		}

		return parts;
	}

	public String toString() {
		return pattern;
	}
}
