package de.bioinf.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.net.URL;
import java.util.zip.GZIPInputStream;

/**
 * A class for reading text lines from files or URL's. Additionally when invoking LineReader(String,boolean) a message
 * is send to Logger.info().
 * 
 * <ul>
 * <li>The method putBack() is for rereading the last line of a file. This may be useful if a line must be evaluated and
 * read in a next call again.</li>
 * <li>Depending on the filename (or url's name) even g-zipped files may be read (automatically).</li>
 * </ul>
 * 
 * @author ng
 */
public class LineReader {
	private String filename = null;
	private BufferedReader reader = null;
	/**
	 * The line number
	 */
	private int lineno = 0;
	/**
	 * The last read line
	 */
	private String line = null;
	/**
	 * Put the last read line back to read it again.
	 */
	private boolean putBack = false;
	private InfoThread infoThread = null;

	/**
	 * A thread for writing loading progress info
	 */
	class InfoThread extends ProgressThread {
		private long fsize = 0;
		private long size = 0;

		public InfoThread(long fsize) {
			super(1);
			this.fsize = fsize;
		}

		public void addSize(long size) {
			this.size += size;
		}

		@Override
		public String getMessage() {
			if (fsize > 0) {
				return String.format("Loading '%s' [%.2f %%]", filename, (double) size / fsize * 100.0);
			} else {
				return String.format("Loading '%s' [%d bytes]", filename, size);
			}
		}
	}

	public LineReader(File file, boolean progress) throws BioinfException {
		super();
		try {
			init(file.getAbsolutePath(), new FileInputStream(file), progress, file.length());
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
	}

	public LineReader(File file) throws BioinfException {
		this(file, false);
	}

	public LineReader(String filename, boolean progress) throws BioinfException {
		this(new File(filename), progress);
	}

	public LineReader(String filename) throws BioinfException {
		this(filename, false);
	}

	public LineReader(URL url, boolean progress) throws BioinfException {
		super();
		try {
			init(url.getFile(), url.openStream(), progress, url.openConnection().getContentLength());
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
	}

	public LineReader(URL url) throws BioinfException {
		this(url, false);
	}

	public LineReader(Reader reader) throws BioinfException {
		super();
		try {
			this.reader = new BufferedReader(reader);
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage("Error opening reader!");
		}
	}

	public String getFilename() {
		return filename;
	}

	public String readLine() throws BioinfException {
		if (putBack) {
			putBack = false;
			return line;
		}
		try {
			line = reader.readLine();
			lineno++;
			if (infoThread != null && line != null) {
				infoThread.addSize(line.length());
			}
			return line;
		} catch (Exception e) {
			throw new BioinfException(String.format("Error reading line %d from file %s!", lineno, filename));
		}
	}

	public void close() throws BioinfException {
		try {
			if (infoThread != null) {
				infoThread.terminate();
			}
			reader.close();
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(String.format("Error closing file %s!", filename));
		}
	}

	/**
	 * Returns the last read line number. Line number 0 means no line is read so far.
	 */
	public int getLineNumber() {
		return lineno;
	}

	/**
	 * Put the last read line back to read it gain.
	 */
	public void putBack() {
		if (line != null) {
			putBack = true;
		}
	}

	/**
	 * Read and return size characters without positioning the file pointer. If the file is shorter than size, the
	 * returned array will fit to the actual file size.
	 */
	public char[] readAhead(int size) throws BioinfException {
		BioinfException.fire(reader.markSupported(), "Cannot read ahead!");
		try {
			char[] a = new char[size];
			reader.mark(a.length);
			int count = reader.read(a);
			reader.reset();
			return count < a.length ? new String(a, 0, count).toCharArray() : a;
		} catch (Exception e) {
			throw BioinfException.convert(e);
		}
	}

	// **
	// Private
	// **

	private void init(String filename, InputStream stream, boolean progress, long size) throws BioinfException {
		this.filename = filename;
		try {
			reader = new BufferedReader(new InputStreamReader(getStream(stream)));
			check();
			if (progress) {
				infoThread = new InfoThread(stream instanceof GZIPInputStream ? 0 : size);
				infoThread.start();
			}
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(String.format("Error opening %s!", filename));
		}
	}

	private InputStream getStream(InputStream stream) throws BioinfException {
		try {
			byte[] sign = new byte[2];
			PushbackInputStream pbstream = new PushbackInputStream(stream, sign.length);
			pbstream.read(sign);
			pbstream.unread(sign);
			return sign[0] == 0x1f && sign[1] == 0x8b ? new GZIPInputStream(pbstream) : pbstream;
		} catch (Exception e) {
			throw new BioinfException(e);
		}
	}

	private void check() throws BioinfException {
		for (char c : readAhead(1024 * 48)) {
			if (Character.isISOControl(c) && "\n\r\t".indexOf(c) < 0) {
				throw new BioinfException("Invalid character in file!");
			}
		}
	}
}
