package de.bioinf.utils;

import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;

/**
 * A class (used similar as a singleton) to log messages. The info() and warning() methods forward messages to a
 * listener (see also addListener()), they will not be printed into a file. The error() and trace() methods write
 * messages to a file (see redirect()) or the standard output if not redirected.
 * 
 * @author ng
 */
public class Logger {
	private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy.MM.dd HH:mm:ss");

	private static ArrayList<MessageListener> listeners = new ArrayList<MessageListener>();
	private static boolean trace = false;
	private static PrintWriter logstream = new PrintWriter(System.out);
	private static boolean log2info = false;

	private Logger() {
	}

	/**
	 * Appends coming messages to the specified file instead of using the standard output.
	 */
	public static void redirect(String filename) throws BioinfException {
		try {
			logstream = new PrintWriter(new FileOutputStream(filename, true));
			logstream.println(String.format("=== [%s] ===", Utils.getCurrentTime()));
		} catch (Exception e) {
			throw BioinfException.convert(e).addMessage(String.format("Cannot redirect logging to '%s'!", filename));
		}
	}

	/**
	 * Additional redirect all errors to info().
	 */
	public static void log2info() {
		log2info = true;
	}

	public static void addListener(MessageListener listener) {
		listeners.add(listener);
	}

	public static void removeListener(MessageListener listener) {
		listeners.remove(listener);
	}

	public static void trace(String message) {
		log(MessageEvent.TRACE, message, null, null);
	}

	public static void error(String message) {
		log(MessageEvent.ERROR, message, null, null);
	}

	public static void error(Exception e) {
		log(MessageEvent.ERROR, e.getMessage(), null, null);
	}

	public static void error(String message, Exception e) {
		log(MessageEvent.ERROR, message, null, e);
	}

	public static void error(String message, Object o, Exception e) {
		log(MessageEvent.ERROR, message, o, e);
	}

	public static void info(String message) {
		callListeners(MessageEvent.INFO, message);
	}

	/**
	 * If the condition is false write an info message and return the condition.
	 */
	public static boolean info(boolean cond, String message) {
		if (!cond) {
			info(message);
		}
		return cond;
	}

	/**
	 * Sets trace on or off
	 */
	public static void setTrace(boolean state) {
		trace = state;
	}

	/**
	 * Is trace mode on?
	 */
	public static boolean isTrace() {
		return trace;
	}

	// **
	// Private
	// **

	private static void log(int type, String message, Object o, Exception e) {
		try {
			StringBuffer buf = new StringBuffer(128);
			buf.append("[");
			buf.append(toString(type)).append(" / ");
			buf.append(DATE_FORMAT.format(new Date(System.currentTimeMillis())));
			if (o != null) {
				buf.append(Utils.getName(o)).append(" / ");
			}
			buf.append("] ");
			buf.append(message);
			if (e != null) {
				buf.append(" (").append(e.getMessage()).append(")");
			}
			logstream.println(buf.toString());
			if (e != null) {
				e.printStackTrace(logstream);
			}
			logstream.flush();
			if (log2info || type != MessageEvent.TRACE) {
				if (e != null) {
					message = String.format("%s (%s)", message, e.getMessage());
				}
				callListeners(type, message);
			}
		} catch (Exception e2) {
			System.err.println(e2);
		}
	}

	private static void callListeners(int type, String message) {
		for (MessageListener l : listeners) {
			l.notice(new MessageEvent(message, type));
		}
	}

	private static String toString(int type) {
		switch (type) {
		case MessageEvent.TRACE:
			return "TRACE";
		case MessageEvent.INFO:
			return "INFO";
		case MessageEvent.ERROR:
			return "ERR";
		default:
			return "???";
		}
	}
}
