package de.bioinf.utils;

/**
 * A 1-dimensional value range. It is possible to take care of adding values to a coordinate system for calculating the
 * maximum borders/limits of the inserted values.
 * 
 * @author ng
 */
public class ValueRange {
	private double min = Double.MAX_VALUE;
	private double max = Double.MIN_VALUE;

	public ValueRange() {
		super();
	}

	public ValueRange(double min, double max) {
		super();
		set(min, max);
	}

	public ValueRange(IntValueRange range) {
		super();
		set(range.getMinVal(), range.getMaxVal());
	}

	/**
	 * "Add" a value and calculate the limits anew.
	 */
	public void add(double value) {
		min = Math.min(min, value);
		max = Math.max(max, value);
	}

	/**
	 * "Add" a range and calculate the limits anew.
	 */
	public void add(ValueRange range) {
		if (range.isValid()) {
			min = Math.min(min, range.min);
			max = Math.max(max, range.max);
		}
	}

	public boolean equals(Object o) {
		if (o == null || !(o instanceof ValueRange)) {
			return false;
		}
		ValueRange r = (ValueRange) o;
		return this.min == r.min && this.max == r.max;
	}

	/**
	 * Returns the minimum of all added values.
	 * <p>
	 * Remark: If nothing is added Double.MAX_VALUE will be returned!
	 */
	public double getMinVal() {
		return min;
	}

	/**
	 * Returns the maximum of all added values.
	 * <p>
	 * Remark: If nothing is added Double.MIN_VALUE will be returned!
	 */
	public double getMaxVal() {
		return max;
	}

	/**
	 * Returns a set of values between min and max. This values can be taken as a coordinate x/y axis value for line
	 * drawing. The values are that way, so they have as less as possible digits.
	 * 
	 * @param approx - the approximately count of returned values.
	 */
	public double[] getIntermediaryValues(int approx) {
		if (!isValid()) {
			return new double[0];
		}
		double diff = max - min;
		double step = 1;
		if (diff > 1) {
			while ((int) (diff / step) > 9) {
				step *= 10.0;
			}
		} else {
			while ((int) (diff * step) < 1) {
				step *= 10.0;
			}
			step = 1 / step;
		}
		while (diff / step < approx) {
			if (diff * 2 / step >= approx) {
				step /= 2;
			} else if (diff * 4 / step >= approx) {
				step /= 4;
			} else if (diff * 5 / step >= approx) {
				step /= 5;
			} else {
				step /= 10;
			}
		}
		double start = (int) (min / step + (min <= 0 ? 0.0 : 1.0)) * step;
		if (start == min) {
			start += step;
		}
		int count = 0;
		while (start + step * count < max) {
			count++;
		}
		double res[] = new double[count];
		for (int i = 0; i < count; i++) {
			res[i] = Math.round(start * 1000000000.0) / 1000000000.0;
			start += step;
		}
		return res;
	}

	/**
	 * Returns the difference between the maximum (added) value and the minimum (added) value.
	 */
	public double getRange() {
		return max - min;
	}

	public void init() {
		min = Double.MAX_VALUE;
		max = Double.MIN_VALUE;
	}

	/**
	 * Do the 2 ranges intersect?
	 */
	public boolean intersects(ValueRange range) {
		return isInside(range.min) || range.isInside(min);
	}

	/**
	 * Is the value inside the range (including min/max)?
	 */
	public boolean isInside(double value) {
		return min <= value && value <= max;
	}

	/**
	 * Is this a valid range (max > min)?
	 */
	public boolean isValid() {
		return min != Double.NaN && max != Double.NaN && max > min;
	}

	/**
	 * Set the limts.
	 */
	public void set(double min, double max) {
		this.min = min;
		this.max = max;
	}

	public String toString() {
		return "[" + min + ", " + max + "]";
	}
}
